"""The clock manager subsystem timestamps readings and implements timers.

No other controller subsystem has a concept of device time.  They are all
purely reactive to whatever work items are queued for them.  This is
by design since IOTile devices are meant to be event-driven rather than
polling based whenever possible.

However, there are times when you need to queue operations to be performed
periodically.  It is the clock manager subsystem's job to allow for queuing
these periodic tasks.  Internally, all timers work by just sending a
periodic input to the sensor graph subsystem.  There are 4 timer inputs
available:

- normal tick: generated every 10 seconds in all circumstances.
- fast tick: generated every 1 second when enabled.
- user tick 1: generated at a user configurable interval and easily
  updated dynamically using an RPC.
- user tick 2: generated at a user configurable interval and easily
  updated dynamically using an RPC.

The first two ticks are reserved for internal use and cannot be adjusted
by the user, hower the last two ticks are designed for easy control in
the field.


Device Time
===========

IOTile devices do not require a concept of UTC time in order to function.
Those device that include a properly synchronized realtime-clock (RTC) can
transparently gain the benefits of UTC time stamping on all of their generated
data, however low cost devices that do not have a proper RTC can also be
implemented using the device uptime as the source of timestamps rather than
a UTC value.

Internally, all timestamps in an IOTile device are represented to 1 second
precision using a uint32_t.  UTC timestamps are distinguished from uptime
timestamps by setting the MSB of the timestamps.  The epoch used for UTC
timestamps is Jan 1, 2000 00:00, i.e. the millenium, not the unix epoch.


Emulation Notes
===============

There are many cases when you do not want an emulated device to experience the
passage of time in the same way as a real device.  For example, you may want
to write test cases that inspect the timestamps generated by a device and so
you need control over what the device time will be.  You may also want to
perform a low level logic test where you verify what events will occur when a
specific timer input is sent to the sensor graph subsystem.  In this case
it is particularly important that you not have non-deterministic other inputs
being sent to sensor-graph depending on how long the test took to execute.

There are other cases where you want to run an end-to-end type test and want
to see the periodic data produced by an emulated device after X minutes or
hours have passed.

To handle both of these use cases, the emulated clock subsystem supports running
a device with time enable or disabled.  If time is disabled, the device's uptime
is frozen at a fixed value and no periodic inputs are generated to sensor-graph.

You are free to send the same periodic inputs to sensor-graph yourself using
rpcs if you wish to control the passage of emulation time (as seen by your
sensor-graph rules) in a fine-grained fashion.

Handling Time When Loading Snapshots
-----------------------------------

TODO: figure out a good set of options for resuming a device when utc time
      is set.
"""

from iotile.core.hw.virtual import tile_rpc
from ...constants import rpcs, pack_error, Error, ControllerSubsystem, streams
from .controller_system import ControllerSubsystemBase


class ClockManagerSubsystem(ControllerSubsystemBase):
    """Container state of the clock manager subsystem."""

    FAST_TICK = 0
    USER1_TICK = 1
    USER2_TICK = 2

    TICK_NAMES = {
        FAST_TICK: 'fast',
        USER1_TICK: 'user1',
        USER2_TICK: 'user2'
    }

    TICK_STREAMS = {
        'fast': streams.FAST_TICK,
        'normal': streams.NORMAL_TICK,
        'user1': streams.USER_TICK_1,
        'user2': streams.USER_TICK_2
    }

    def __init__(self, emulator, has_rtc=False):
        super(ClockManagerSubsystem, self).__init__(emulator)

        self.uptime = 0
        self.time_offset = 0
        self.is_utc = False

        self.stored_offset = None
        self.has_rtc = has_rtc
        self.ticks = dict(fast=0, user1=0, user2=0, normal=10)
        self.tick_counters = dict(fast=0, user1=0, user2=0, normal=0)

        # Hook for allowing us to link this subsystem to sensor_graph
        self.graph_input = lambda x, y: None

    def clear_to_reset(self, config_vars):
        """Clear all volatile information across a reset.

        The reset behavior is that:
        - uptime is reset to 0
        - is `has_rtc` is True, the utc_offset is preserved
        - otherwise the utc_offset is cleared to none
        """

        super(ClockManagerSubsystem, self).clear_to_reset(config_vars)

        self.tick_counters = dict(fast=0, user1=0, user2=0, normal=0)

        self.is_utc = False
        self.time_offset = 0

        if self.has_rtc and self.stored_offset is not None:
            self.time_offset = self.stored_offset + self.uptime

        self.uptime = 0

        self.ticks['fast'] = config_vars.get('fast_tick', 0)
        self.ticks['user1'] = config_vars.get('user_tick_1', 0)
        self.ticks['user2'] = config_vars.get('user_tick_2', 0)

    def tick_name(self, tick_index):
        """Convert a tick index into a string name."""

        return self.TICK_NAMES.get(tick_index)

    def handle_tick(self):
        """Internal callback every time 1 second has passed."""

        self.uptime += 1

        for name, interval in self.ticks.items():
            if interval == 0:
                continue

            self.tick_counters[name] += 1
            if self.tick_counters[name] == interval:
                self.graph_input(self.TICK_STREAMS[name], self.uptime)
                self.tick_counters[name] = 0

    def set_tick(self, index, interval):
        """Update the a tick's interval.

        Args:
            index (int): The index of the tick that you want to fetch.
            interval (int): The number of seconds between ticks.
                Setting this to 0 will disable the tick.

        Returns:
            int: An error code.
        """

        name = self.tick_name(index)
        if name is None:
            return pack_error(ControllerSubsystem.SENSOR_GRAPH, Error.INVALID_ARRAY_KEY)

        self.ticks[name] = interval
        return Error.NO_ERROR

    def get_tick(self, index):
        """Get a tick's interval.

        Args:
            index (int): The index of the tick that you want to fetch.

        Returns:
            int, int: Error code and The tick's interval in seconds.

            A value of 0 means that the tick is disabled.
        """

        name = self.tick_name(index)
        if name is None:
            return [pack_error(ControllerSubsystem.SENSOR_GRAPH, Error.INVALID_ARRAY_KEY), 0]

        return [Error.NO_ERROR, self.ticks[name]]

    def get_time(self, force_uptime=False):
        """Get the current UTC time or uptime.

        By default, this method will return UTC time if possible and fall back
        to uptime if not.  If you specify, force_uptime=True, it will always
        return uptime even if utc time is available.

        Args:
            force_uptime (bool): Always return uptime, defaults to False.

        Returns:
            int: The current uptime or encoded utc time.
        """

        if force_uptime:
            return self.uptime

        time = self.uptime + self.time_offset

        if self.is_utc:
            time |= (1 << 31)

        return time

    def synchronize_clock(self, offset):
        """Persistently synchronize the clock to UTC time.

        Args:
            offset (int): The number of seconds since 1/1/2000 00:00Z
        """

        self.time_offset = offset - self.uptime
        self.is_utc = True

        if self.has_rtc:
            self.stored_offset = self.time_offset


class ClockManagerMixin(object):
    """Mixin to add the clock manager subsystem into ReferenceController."""

    def __init__(self, emulator, has_rtc):
        self.clock_manager = ClockManagerSubsystem(emulator, has_rtc)

        self.declare_config_variable('fast_tick', 0x2000, 'uint32_t', default=0)
        self.declare_config_variable('user_tick_1', 0x2002, 'uint32_t', default=0)
        self.declare_config_variable('user_tick_2', 0x2003, 'uint32_t', default=0)

        self._post_config_subsystems.append(self.clock_manager)

    @tile_rpc(*rpcs.GET_USER_TIMER)
    def get_user_timer(self, index):
        """Get the current value of a user timer."""

        err, tick = self.clock_manager.get_tick(index)
        return [err, tick]

    @tile_rpc(*rpcs.SET_USER_TIMER)
    def set_user_timer(self, value, index):
        """Set the current value of a user timer."""

        err = self.clock_manager.set_tick(index, value)
        return [err]

    @tile_rpc(*rpcs.GET_CURRENT_TIME)
    def get_current_time(self):
        """Get the device's current time."""

        time = self.clock_manager.get_time()

        return [time]

    @tile_rpc(*rpcs.GET_CURRENT_UPTIME)
    def get_current_uptime(self):
        """Get the device's current time."""

        time = self.clock_manager.get_time(force_uptime=True)

        return [time]

    @tile_rpc(*rpcs.GET_UTC_TIME_OFFSET)
    def get_time_offset(self):
        """Get the currently programmed time offset."""

        return [self.clock_manager.time_offset, int(self.clock_manager.is_utc)]

    @tile_rpc(*rpcs.SET_UTC_TIME_OFFSET)
    def set_time_offset(self, offset, is_utc):
        """Temporarily set the current time offset."""

        is_utc = bool(is_utc)
        self.clock_manager.time_offset = offset
        self.clock_manager.is_utc = is_utc

        return [Error.NO_ERROR]

    @tile_rpc(*rpcs.SYNCHRONIZE_CLOCK)
    def synchronize_clock(self, offset):
        """Persistently synchronize the device's clock to UTC."""

        self.clock_manager.synchronize_clock(offset)

        return [Error.NO_ERROR]



"""A simple emulated reference device that includes a demo non-controller tile.

The demo emulated Peripheral tile showcases the 4 kinds of things that you can
do on an EmulatedPeripheralTile:

- Normal synchronous RPCS: These are methods decorated with the
  @tile_rpc decorator and directly invocable using the `rpc` method on an
  EmulatedDevice or via a proxy class.  The methods are run in the
  EmulationLoop inside of the RPC dispatch task.

  They may not yield since they are not coroutines.  If they block the entire
  emulation loop will block until they finish.  For that reason, they should
  not block.

- Asynchronous RPCS: These are RPCs whose implementation requires that they
  return their result via a callback.  In physical IOTile devices this is
  usually because the operation requires an unbounded amount of time to
  complete, such as asking an external sensor for a value that could take many
  ms to respond.

  These RPCs are typically implemented by queuing work for a background task
  and then raising AsynchronousRPCResponse().  This notifies the RPC dispatch
  that the RPC will be returning its response later and allows for other RPCs
  to be dispatched in the meantime.

  A background worker task associated with this tile can then call
  EmulatedDevice.finish_async_rpc when it wants to complete the RPC call.

  From the point of view of the caller of the RPC, it will block until
  finish_async_rpc has been called with the response and it does not have any
  way to know whether the response was given immediately or via a callback to
  finish_async_rpc.  Put another way, RPCs are always synchronous from the
  caller's perspective.  It is just that sometimes you don't want to block the
  RPC dispatch during a long-running RPC so that other callers can send RPCs
  while your caller is still waiting for its response.

- Background Loops: In a physical tile, there is a single main loop and
  RPC handlers are invoked as interrupts.  In EmulatedPeripheralTile subclasses,
  the _application_main() coroutine is the equivalent of the main loop.

  It is launched when the tile starts and should loop forever.  It is canceled
  clenaly when the tile is reset and started again.  If you need to perform
  tasks periodically that are not related to any specific RPC, they should
  go inside your _application_main().  Similarly, if you support asynchronous
  rpcs, they should be implemented by queuing work that is processed by
  _application_main().

- Coroutine RPCs: For some RPCs it may be easier to implement them if they can
  await an object.  In this case the RPC never raises AsynchronousRPCResponse
  and still blocks the RPC dispatch thread until it finishes, it is just that
  you are allowed to `await` an object inside your implementation.  These RPCs
  must be decorated with `@async_tile_rpc` instead of `@tile_rpc`.

The DemoEmulatedTile class in this module shows an example of how to perform
each of these 4 things.  Note that for ease of implementation, peripheral
tiles are able to directly stream and trace data out of the EmulatedDevice
rather than needing to invoke an RPC on the controller to stream or trace on
their behalf.
"""

import asyncio
import logging
from iotile.core.hw.virtual import tile_rpc
from iotile.core.hw.virtual.common_types import AsynchronousRPCResponse, pack_rpc_payload
from ..virtual import EmulatedPeripheralTile
from ..internal import async_tile_rpc
from ..reference import ReferenceDevice


class DemoEmulatedTile(EmulatedPeripheralTile):
    """A basic emulated tile showing all of the things you can do.

    This demo tile has a trivial main loop that pulls work queued by RPCs and
    executes it.  It has several synchronous RPCs, one asynchronous RPC, an
    RPC implemented as a coroutine and an RPC that triggers the tile to trace
    a bunch of binary data.
    """

    name = b'emudmo'

    def __init__(self, address, device):
        super(DemoEmulatedTile, self).__init__(address, device)
        self.register_scenario('loaded_counter', self.load_counter)
        self._counter = 0
        self._work = device.emulator.create_queue(register=False)
        self._logger = logging.getLogger(__name__)

    async def _application_main(self):
        self.initialized.set()

        while True:
            action, args = await self._work.get()

            try:
                if action == 'echo':
                    rpc_id, arg = args
                    self._device.emulator.finish_async_rpc(self.address, rpc_id, pack_rpc_payload("L", (arg,)))
                elif action == 'trace':
                    byte_count = args
                    chunks = byte_count // 20
                    if byte_count % 20:
                        chunks += 1

                    for i in range(0, chunks):
                        chunk_length = min(byte_count - i*20, 20)
                        data = bytes(range(0, chunk_length))

                        self._logger.debug("Tracing chunk %d/%d (size=%d)", i + 1, chunks, chunk_length)

                        success = await self._device.trace_sync(data)
                        if not success:
                            self._logger.error("Failure sending chunk %d, aborting", i)
                            break

                    if success:
                        self._logger.info("Finished sending %d chunks of tracing data", chunks)
                else:
                    self._logger.error("Unknown action in main loop: %s", action)
            except:
                self._logger.exception("Error processing background action: action=%s, args=%s", action, args)

    def load_counter(self, counter):
        """Load the counter value of this device."""
        self._counter = counter

    @tile_rpc(0x8000, "L", "L")
    def async_echo(self, arg):
        """Asynchronously echo the argument number."""

        self._work.put_nowait(('echo', (0x8000, arg)))
        raise AsynchronousRPCResponse()

    @tile_rpc(0x8001, "L", "L")
    def sync_echo(self, arg):
        """Synchronously echo the argument number."""

        return [arg]

    @tile_rpc(0x8002, "", "L")
    def counter(self):
        """A counter that increments everytime it is called."""

        value = self._counter
        self._counter += 1
        return [value]

    @tile_rpc(0x8003, "L", "")
    def start_trace(self, count):
        """Start tracing a given number of bytes."""

        self._work.put_nowait(('trace', count))
        return []

    @async_tile_rpc(0x8004, "L", "L")
    async def coroutine_echo(self, arg):
        """Wait for a small delay and then echo.

        This method shows an example of how to implement an RPC as a
        coroutine.
        """

        await asyncio.sleep(0.01)
        return [arg]


class DemoEmulatedDevice(ReferenceDevice):
    """A basic emulated device that includes a single blank tile in addition to the reference controller.

    The blank tile in slot 1 has module name emudmo and the following two config variables declared
    to allow for testing config variable usage and streaming:

    - 0x8000: uint32_t
    - 0x8001: uint8_t[16]

    Args:
        args (dict): A dictionary of optional creation arguments.  Currently
            supported are:
                iotile_id (int or hex string): The id of this device. This
                defaults to 1 if not specified.
    """

    STATE_NAME = "emulation_demo_device"
    STATE_VERSION = "0.1.0"

    def __init__(self, args):
        super(DemoEmulatedDevice, self).__init__(args)

        peripheral = DemoEmulatedTile(11, device=self)
        peripheral.declare_config_variable('test 1', 0x8000, 'uint32_t')
        peripheral.declare_config_variable('test 2', 0x8001, 'uint8_t[16]')

        self.add_tile(11, peripheral)



"""Constants related to the TileManager subsystem on an IOTile controller."""

from enum import IntEnum

class RunLevel(IntEnum):
    """Possible directives returned to a tile upon registration.

    These run level directives tell the tile if and when it should pass
    control to application firmware.  A tile's run level is assigned exactly
    once when it resets itself and calls the REGISTER_TILE rpc.

    The run level is then fixed until the tile is reset at which point a new
    run level can be received from the TileBus controller.
    """

    START_IMMEDIATELY = 0
    """Pass control to application firmware immediately.

    This run level is not used inside physical IOTile devices because it does
    not provide the controller with time to safely stream configuration
    information to the POD before the application firmware runs and could
    potentially reference the config variables.

    The START_IMMEDIATELY run level exists to allow for unit testing scenarios
    when tiles should run independently of a controller.
    """

    START_ON_COMMAND = 1
    """Wait until a RUN_APPLICATION rpc is received before passing control to app.

    This is the standard run level returned to a physical tile.  It tells the
    tile's executive to wait until the controller sends it a RUN_APPLICATION
    rpc, at which point it can pass control to the any loaded application
    firmware.
    """

    SAFE_MODE = 2
    """The tile should enter safe mode, which means no application control.

    Safe mode tells the tile executive that the application firmware is
    potentially misbehaving and should not be run.  This prevents
    loss-of-control situations where a misbehaving tile takes control of
    TileBus and prevents the receipt of any messages telling it to stop
    talking.

    Since the controller can cut power to all tiles and the tile executive
    waits before running application firmware, there is always a de-bricking
    mechanism by cutting power to the tile and then telling it to enter safe
    mode, whereby executive RPC functions can be executed including
    bootloading.
    """

    DORMANT = 3
    """The tile should enter a deep sleep state until reset by the controller.

    This state is not currently used but is designed to allow infrequently
    used tiles to be put into an ultra-low-power state that does not preserve
    RAM contents or allow for processing of RPCs.
    """

    BOOTLOAD = 4
    """The tile should immediately bootload firmware from the controller.

    This starts a tile's firmware upgrade process where it passes control to
    its bootloader that begins pulling firmware from the controller and
    updating itself.
    """


class TileState(IntEnum):
    """Possible states that a tile can have inside the TileManager cache.

    These states represent the phase of the tile lifecycle that the tile
    is currently known to be in.
    """

    INVALID = 0
    """This slot in the tile manager cache does not represent a tile."""

    JUST_REGISTERED = 1
    """The tile has just sent the REGISTER_TILE rpc.

    This state changes to BEING_CONFIGURED when the controller starts sending
    config variables to the tile.
    """

    BEING_CONFIGURED = 2
    """The tile is currently receiving config variables from the controller."""

    RUNNING = 3
    """The tile has received the START_APPLICATION rpc from the controller."""

    SAFE_MODE = 4
    """The controller replied to the tile's REGISTER_TILE rpc with RunLevel.SAFE_MODE."""

    DORMANT = 5
    """The controller replied to the tile's REGISTER_TILE rpc with RunLevel.DORMANT."""

    SHOULD_BOOTLOAD = 6
    """The controller replied to the tile's REGISTER_TILE rpc with RunLevel.BOOTLOAD."""

    BOOTLOADING = 7
    """The tile was told to bootload and it subsequently reported that it had started."""

    FINISHED_BOOTLOADING = 8
    """The tile was told to bootload and it subsequently reported that it had finished."""

    UNRECOVERABLE_ERROR = 9
    """The tile has performed an action that indicates it had a fatal error."""



import binascii
import struct
from iotile.core.hw.virtual.common_types import RPCInvalidIDError, unpack_rpc_payload, pack_rpc_payload, RPCInvalidArgumentsError, RPCInvalidReturnValueError


def async_rpc(address, rpc_id, arg_format, resp_format=None):
    """Decorator to denote that a function implements an RPC with the given ID and address.

    The underlying function should be a member function that will take
    individual parameters after the RPC payload has been decoded according
    to arg_format.

    Arguments to the function are decoded from the 20 byte RPC argument payload according
    to arg_format, which should be a format string that can be passed to struct.unpack.

    Similarly, the function being decorated should return an iterable of results that
    will be encoded into a 20 byte response buffer by struct.pack using resp_format as
    the format string.

    The RPC will respond as if it were implemented by a tile at address ``address`` and
    the 16-bit RPC id ``rpc_id``.

    Args:
        address (int): The address of the mock tile this RPC is for
        rpc_id (int): The number of the RPC
        arg_format (string): a struct format code (without the <) for the
            parameter format for this RPC.  This format code may include the final
            character V, which means that it expects a variable length bytearray.
        resp_format (string): an optional format code (without the <) for
            the response format for this RPC. This format code may include the final
            character V, which means that it expects a variable length bytearray.
    """

    if rpc_id < 0 or rpc_id > 0xFFFF:
        raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

    def _rpc_wrapper(func):
        async def _rpc_executor(self, payload):
            try:
                args = unpack_rpc_payload(arg_format, payload)
            except struct.error as exc:
                raise RPCInvalidArgumentsError(str(exc), arg_format=arg_format, payload=binascii.hexlify(payload))

            resp = await func(self, *args)

            if resp is None:
                resp = []

            if resp_format is not None:
                try:
                    return pack_rpc_payload(resp_format, resp)
                except struct.error as exc:
                    raise RPCInvalidReturnValueError(str(exc), resp_format=resp_format, resp=repr(resp))

            return resp

        _rpc_executor.rpc_id = rpc_id
        _rpc_executor.rpc_addr = address
        _rpc_executor.is_rpc = True
        return _rpc_executor

    return _rpc_wrapper


def async_tile_rpc(rpc_id, arg_format, resp_format=None):
    """Decorator to denote that a function implements an RPC with the given ID on a tile.

    The underlying function should be a member function that will take
    individual parameters after the RPC payload has been decoded according
    to arg_format.

    Arguments to the function are decoded from the 20 byte RPC argument payload according
    to arg_format, which should be a format string that can be passed to struct.unpack.

    Similarly, the function being decorated should return an iterable of results that
    will be encoded into a 20 byte response buffer by struct.pack using resp_format as
    the format string.

    The RPC will respond as if it were implemented by a tile at address ``address`` and
    the 16-bit RPC id ``rpc_id``.

    Args:
        rpc_id (int): The number of the RPC
        arg_format (string): a struct format code (without the <) for the
            parameter format for this RPC
        resp_format (string): an optional format code (without the <) for
            the response format for this RPC
    """

    return async_rpc(None, rpc_id, arg_format, resp_format)



"""A simply proxy class for controlling the DemoEmulatedTile."""

from typedargs.annotate import docannotate, context
from iotile.core.hw.proxy import TileBusProxyObject
from iotile.core.exceptions import ArgumentError

@context("DemoTile")
class DemoTileProxy(TileBusProxyObject):
    @classmethod
    def ModuleName(cls):
        return 'emudmo'

    @docannotate
    def trace_data(self, byte_count):
        """Trigger the tracing of data.

        This method lets you test receiving traced data out of a device. It
        will synchronously wait unitl the given number of bytes have been
        received.

        Args:
            byte_count (int): The number of bytes that should be traced.

        Returns:
            int: The number of bytes that were successfully received.
        """

        self.rpc_v2(0x8003, "L", "", byte_count)

        data = self._hwmanager.wait_trace(byte_count, timeout=((.15 * ((byte_count // 20) + 1)) + .5))
        return len(data)

    @docannotate
    def fetch_counter(self):
        """Fetch and increment the tile's counter value.

        This method returns the current counter and causes the tile
        to increment it by one.

        Returns:
            int: The current counter value before incrementing.
        """

        data, = self.rpc_v2(0x8002, "", "L")
        return data

    @docannotate
    def echo(self, value, method="sync"):
        """Echo a number.

        The echo RPC on the emulated tile are implemented using
        three different methods that you can choose by passing
        method=(sync|async|coroutine)

        The result of all three methods is the same from the caller's
        point of view.

        Args:
            value (int): The value to echo
            method (str): The specific RPC implementation to use.  This
                may be sync, async or coroutine.

        Returns:
            int: The echoed value.
        """

        rpc_map = {
            'sync': 0x8001,
            'async': 0x8000,
            'coroutine': 0x8004
        }

        rpc_id = rpc_map.get(method)
        if rpc_id is None:
            raise ArgumentError("Unknown method: %s" % method, known_methods=list(rpc_map))

        response, = self.rpc_v2(rpc_id, "L", "L", value)
        return response



"""Test coverage of the sensorgraph subsystem on the reference controller."""

import sys
import pytest
import time
from iotile.core.hw import HardwareManager
from iotile.core.hw.reports import SignedListReport
from iotile.core.exceptions import HardwareError
from iotile.core.hw.proxy.external_proxy import find_proxy_plugin
from iotile.emulate.virtual import EmulatedPeripheralTile
from iotile.emulate.reference import ReferenceDevice
from iotile.emulate.demo import DemoEmulatedDevice
from iotile.emulate.constants import rpcs, Error
from iotile.emulate.transport import EmulatedDeviceAdapter

@pytest.fixture(scope="function")
def sg_device():
    """Get a reference device and connected HardwareManager."""

    device = ReferenceDevice({'simulate_time': False})
    peripheral = EmulatedPeripheralTile(11, device)
    peripheral.declare_config_variable("test 1", 0x8000, 'uint16_t')
    peripheral.declare_config_variable('test 2', 0x8001, 'uint32_t[5]')

    device.add_tile(11, peripheral)

    adapter = EmulatedDeviceAdapter(None, devices=[device])

    with HardwareManager(adapter=adapter) as hw:
        hw.connect(1)

        con = hw.get(8, basic=True)
        sensor_graph = find_proxy_plugin('iotile_standard_library/lib_controller', 'SensorGraphPlugin')(con)

        yield sensor_graph, hw


@pytest.fixture(scope="function")
def basic_sg():
    """A preprogrammed basic sensorgraph for testing."""

    device = ReferenceDevice({'simulate_time': False})
    peripheral = EmulatedPeripheralTile(11, device)
    peripheral.declare_config_variable("test 1", 0x8000, 'uint16_t')
    peripheral.declare_config_variable('test 2', 0x8001, 'uint32_t[5]')

    device.add_tile(11, peripheral)

    adapter = EmulatedDeviceAdapter(None, devices=[device])

    nodes = [
        "(input 1 always) => counter 1024 using copy_latest_a",
        "(counter 1024 when count >= 1 && constant 1030 when value == 1) => counter 1030 using copy_latest_a",
        "(counter 1030 when count >= 4) => output 1 using copy_all_a"
    ]

    with HardwareManager(adapter=adapter) as hw:
        hw.connect(1)

        con = hw.get(8, basic=True)
        sensor_graph = find_proxy_plugin('iotile_standard_library/lib_controller', 'SensorGraphPlugin')(con)

        for node in nodes:
            sensor_graph.add_node(node)

        sensor_graph.add_streamer('output 10', 'controller', True, 'individual', 'telegram')

        yield sensor_graph, hw, device


@pytest.fixture(scope="function")
def streaming_sg():
    """A preprogrammed basic sensorgraph for testing streaming."""

    device = ReferenceDevice({'simulate_time': False})

    adapter = EmulatedDeviceAdapter(None, devices=[device])

    nodes = [
        "(input 1 when count >= 1) => output 1 using copy_latest_a",
        "(input 2 when count >= 1) => output 2 using copy_latest_a",
        "(input 3 when count >= 1 && constant 1 always) => unbuffered 1 using trigger_streamer"
    ]

    with HardwareManager(adapter=adapter) as hw:
        hw.connect(1)

        con = hw.get(8, basic=True)
        sensor_graph = find_proxy_plugin('iotile_standard_library/lib_controller', 'SensorGraphPlugin')(con)

        for node in nodes:
            sensor_graph.add_node(node)

        sensor_graph.push_reading('constant 1', 0)

        sensor_graph.add_streamer('output 1', 'controller', False, 'individual', 'telegram')
        sensor_graph.add_streamer('output 2', 'controller', False, 'hashedlist', 'telegram', withother=0)

        yield sensor_graph, hw, device


@pytest.fixture(scope="function")
def empty_sg():
    """An empty sensorgraph for testing."""

    device = ReferenceDevice({'simulate_time': False})

    adapter = EmulatedDeviceAdapter(None, devices=[device])

    with HardwareManager(adapter=adapter) as hw:
        hw.connect(1)

        con = hw.get(8, basic=True)
        sensor_graph = find_proxy_plugin('iotile_standard_library/lib_controller', 'SensorGraphPlugin')(con)
        yield sensor_graph, hw, device


@pytest.fixture(scope="function")
def rpc_sg():
    """A basic sg for testing the rpc executor."""

    device = DemoEmulatedDevice({'simulate_time': False})

    adapter = EmulatedDeviceAdapter(None, devices=[device])

    nodes = [
        "(input 1 always) => unbuffered 1024 using copy_latest_a",
        "(unbuffered 1024 when count == 1 && constant 1024 always) => output 1 using call_rpc"
    ]

    with HardwareManager(adapter=adapter) as hw:
        hw.connect(1)

        con = hw.get(8, basic=True)
        sensor_graph = find_proxy_plugin('iotile_standard_library/lib_controller', 'SensorGraphPlugin')(con)

        for node in nodes:
            sensor_graph.add_node(node)

        sensor_graph.push_reading('constant 1024', 753666)

        yield sensor_graph, hw, device


def test_inspect_nodes(sg_device):
    """Ensure that we can inspect the sensor graph nodes that we have programmed."""

    nodes = [
        "(input 1 always) => counter 1024 using copy_latest_a",
        "(counter 1024 when count >= 1 && constant 1030 when value == 1) => counter 1030 using copy_latest_a"
    ]

    sg, hw = sg_device

    for node in nodes:
        sg.add_node(node)

    node1 = sg.inspect_node(0)
    node2 = sg.inspect_node(1)

    assert node1 == nodes[0]
    assert node2 == nodes[1]


def test_inspect_streamers(basic_sg):
    """Ensure that we can add and inspect streamers."""

    sg, _hw, _device = basic_sg
    assert str(sg.inspect_streamer(0)) == 'realtime streamer on output 10'


def test_persist(basic_sg):
    """Ensure that sensor_graph persists after reset."""

    sg, hw, _device = basic_sg

    sg.persist()
    assert sg.count_nodes() == 3
    assert str(sg.inspect_streamer(0)) == 'realtime streamer on output 10'

    hw.get(8, basic=True).reset(wait=0)

    # Make sure we recover both streamers and nodes
    assert sg.count_nodes() == 3
    assert str(sg.inspect_streamer(0)) == 'realtime streamer on output 10'


def test_graph_input(basic_sg):
    """Ensure that graph_input works."""

    sg, hw, device = basic_sg
    sg.enable()

    sg.push_reading('constant 1030', 1)
    sg.input('input 1', 1)

    device.wait_idle()

    assert sg.inspect_virtualstream('input 1') == 1
    assert sg.inspect_virtualstream('counter 1024') == 1
    assert sg.count_stream('output 1') == 0

    sg.input('input 1', 2)
    device.wait_idle()
    assert sg.count_stream('output 1') == 0
    sg.input('input 1', 3)
    device.wait_idle()
    assert sg.count_stream('output 1') == 0
    sg.input('input 1', 4)
    device.wait_idle()

    values = sg.download_stream('output 1')
    assert [x.value for x in values] == [4, 4, 4, 4]


def test_streaming(streaming_sg):
    """Ensure that streamers work."""

    sg, hw, _device = streaming_sg
    sg.enable()

    hw.enable_streaming()
    sg.input('input 1', 0)
    sg.input('input 1', 1)
    sg.input('input 1', 2)
    sg.input('input 2', 5)
    sg.input('input 2', 6)
    sg.input('input 2', 7)
    sg.input('input 2', 8)
    sg.input('input 3', 3)

    reports = hw.wait_reports(4, timeout=0.5)
    assert len(reports) == 4
    for i, report in enumerate(reports[:-1]):
        assert report.visible_readings[0].value == i

    signed = reports[-1]
    assert isinstance(signed, SignedListReport)
    readings = signed.visible_readings
    assert len(readings) == 4

    for i, reading in enumerate(readings):
        assert reading.value == 5 + i
        assert reading.reading_id == 5 + i


def test_streaming_multiple_times(empty_sg):
    """Verify that we can send multiple reports from the same streamer.

    Verifies Issue #700.
    """

    sg, hw, _device = empty_sg

    sg.add_streamer('input 1', 'controller', True, 'individual', 'telegram')
    sg.enable()

    hw.enable_streaming()

    sg.input('input 1', 1)
    reports = hw.wait_reports(1, timeout=0.5)
    assert len(reports) == 1
    assert reports[0].visible_readings[0].value == 1

    sg.input('input 1', 2)
    reports = hw.wait_reports(1, timeout=0.5)
    assert len(reports) == 1
    assert reports[0].visible_readings[0].value == 2


def test_rpc_engine(rpc_sg):
    """Ensure that our rpc executor works.

    This test sensorgraph calls an rpc that returns an incrementing counter on
    each call.
    """

    sg, hw, _device = rpc_sg
    sg.enable()

    assert sg.count_readings()['streaming'] == 0

    sg.input('input 1', 0)
    assert sg.count_readings()['streaming'] == 1
    sg.input('input 1', 0)
    assert sg.count_readings()['streaming'] == 2

    values = sg.download_stream('output 1')
    assert [x.value for x in values] == [0, 1]
